/*
 *
 *
 *    AnyRig.ino
 *
 *    Digital USB interface device firmware.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2019 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

// ------------------------- BEGIN CONFIGURATION ------------------------- 

//
//  This firmware has a number of different configuration options.  Some
//  of those configuration options are done in other files. The default
//  values provided should work fine for most purposes.
//
//  One configuration step is REQUIRED -- this is the OUTPUT MODULE
//  configuration, which defines the input and output pin numbers for
//  your hardware.  To configure the output modules, edit 'modules.h',
//  and follow the instructions found there.
//
//  Additional settings can be customized, if desired:
//
//  * To configure the audio codec board, which includes line and
//    microphone input levels, edit the file 'codec.h'.
//
//  * To configure default VOX constants, edit the the file 'vox.h'.
//
//  * To configure default KEY output constants, edit the file 'key.h'.
//
//  * An optional 'clip warning' LED and a CPU 'heartbeat' LED can be
//    configured below.
//
//  Changing other settings is not recommended.
//


//
//  Select the AUDIO CLIP LED pin number; this is a DIGITAL
//    output that drives an optional LED warning of high drive
//    levels that are causing clipping in the audio device.
//
//  Set this to (-1) to disable the feature.
//
#define CLIP_LED_PIN (-1)


//
//  Select the HEARTBEAT LED pin number; this is a DIGITAL
//    output that drives an optional LED that will blink once
//    per second, indicating that the firmware is operating.
//
//  Set this to (-1) to disable the feature.
//
#define HEARTBEAT_LED (20) // D20


//
//  Set the normalized RMS voltage level that will trigger VOX.
//    This sets the VOX sensitivity of the device, and must be
//    a value greater than 0.0, and less than 1.0.  Smaller
//    values make the VOX operation more sensitive.
//
//  This value will apply to both VOX and Key modules.
//
//  The default value of 0.5% is plenty sensitive for both
//    module types, without producing any false triggering.
//
//  TODO: Possibly make this a potentiometer input and/or a
//        runtime/EEPROM configuration option.
//
#define VOX_THRESHOLD (0.005) // 0.5%


//
//  Set this symbol to enable the serial port.
//
//  Enabling this option will allow one or both channels to use
//    DTR and/or RTS keying, in addition to audio-based keying.
//
//  This option requires that the target "USB Type" in the arduino environment
//    be set to one of the combinations that includes both "Serial" and "Audio"
//    such as "Serial + MIDI + Audio".
//
#define ENABLE_SERIAL_KEYING

//
//  OTHER STUFF - don't change these unless you know what you are doing.
//

// how often to read the mixer controls (msec)
#define MIXER_INTERVAL (100) // msec

// clipping constants
#define CLIP_DELAY (250)    // msec; how long to light the CLIP LED

// define the level estimator type, RMS or Peak
#define LevelEstimator AudioAnalyzeRMS // or AudioAnalyzePeak

//
// Set this symbol to enable debugging output to the serial port; enabling this
//    option requires that the target "USB Type" in the arduino environment
//    be set to one of the combinations that includes both "Serial" and "Audio"
//    such as "Serial + MIDI + Audio".
//
//#define VERBOSE_DEBUG

// -------------------------  END  CONFIGURATION ------------------------- 

#include <Audio.h>
//#include <Wire.h>
//#include <SPI.h>
//#include <SD.h>
//#include <SerialFlash.h>
#include "Elapsed.h"
#include "Heartbeat.h"
#include "codec.h"   // this imports the codec configuration
#include "vox.h"     // this imports the VOX module type
#include "key.h"     // this imports the Key module type
#include "modules.h" // this imports the module configuration

//
//  define the software interconnections between the various Teensy DSP/audio modules
//
AudioInputI2S            i2s_in;         //xy=511,192
AudioInputUSB            usb_in;         //xy=511,228
AudioOutputI2S           i2s_out;        //xy=802,224
AudioOutputUSB           usb_out;        //xy=805,188
LevelEstimator           level_left;     //xy=808,282
LevelEstimator           level_right;    //xy=813,316
AudioConnection          patchCord1(i2s_in, 0, usb_out, 0);
AudioConnection          patchCord2(i2s_in, 1, usb_out, 1);
AudioConnection          patchCord3(usb_in, 0, i2s_out, 0);
AudioConnection          patchCord4(usb_in, 0, level_left, 0);
AudioConnection          patchCord5(usb_in, 1, i2s_out, 1);
AudioConnection          patchCord6(usb_in, 1, level_right, 0);
AudioControlSGTL5000     sgtl5000;       //xy=506,320


// the most recent time the mixer level was processed
static uint32_t lastMixer;

// the last time VOX audio clipped
static uint32_t lastClip;

// the most recent threshold states
static bool state_left;
static bool state_right;

// the most recent volume value copied from the soft mixer
static float lastVolume = -1.0;

// create and configure the output modules
static OutputModule *channel1 = ConfigureLeftModule();  // LEFT
static OutputModule *channel2 = ConfigureRightModule(); // RIGHT


//
//  setup()
//
void setup() {
	#if defined (VERBOSE_DEBUG) || defined(ENABLE_SERIAL_KEYING)
	Serial.begin(9600);
	#endif
	#if defined (VERBOSE_DEBUG)
	Serial.println("setup() begin");
	#endif

	// allocate audio buffers
	AudioMemory(12);

	// configure the mixer
	configure_codec(&sgtl5000);

	#if defined (CLIP_LED_PIN) && (CLIP_LED_PIN >= 0)
	// initialize audio LED
	pinMode(CLIP_LED_PIN, OUTPUT);
	digitalWrite(CLIP_LED_PIN, LOW);
	#endif

	// clear flags and timers
	lastMixer = 0;
	lastClip = 0;
	state_left = false;
	state_right = false;

	// setup the keying modules
	if (channel1)
		channel1->Setup();
	if (channel2)
		channel2->Setup();

	// configure the blinking LED
	HeartbeatSetup();

	#ifdef VERBOSE_DEBUG
	Serial.println("setup() end");
	#endif
} // setup()


//
//  loop()
//
void loop() {
	// read the clock
	uint32_t now = millis();

	// blink the LED
	HeartbeatLoop();

	// allow keying modules to update
	if (channel1)
		channel1->Loop();
	if (channel2)
		channel2->Loop();

	//
	//  MIXER: transfer mixer level a few times per second
	//
	if (Elapsed(now, lastMixer) >= MIXER_INTERVAL) {
		lastMixer = now;

		// read the PC's volume setting and clamp it
		float vol = usb_in.volume();
		if (vol > 1.0) vol = 1.0;
		if (vol < 0.0) vol = 0.0;

		// update the codec's volume
		if (vol != lastVolume) {
			lastVolume = vol;
			set_codec_volume(&sgtl5000, vol);
		}
	}

	#if defined (CLIP_LED_PIN) && (CLIP_LED_PIN >= 0)
	bool clip_found = false;
	#endif

	// read RIGHT channel level
	if (level_right.available()) {
		float r = level_right.read();
		state_right = false;
		if (channel2) {
			if (channel2->UseAudio())
				state_right |= (r >= VOX_THRESHOLD);
			#ifdef ENABLE_SERIAL_KEYING
			if (channel2->UseDTR())
				state_right |= (Serial.dtr() ? true : false);
			if (channel2->UseRTS())
				state_right |= (Serial.rts() ? true : false);
			#endif

			#if defined (CLIP_LED_PIN) && (CLIP_LED_PIN >= 0)
			if (r >= channel2->ClipWarning()) {
				clip_found = true;
			}
			#endif

			// update the keying module
			channel2->SetState(state_right);
		}


		#ifdef VERBOSE_DEBUG
		if (r >= 0.01) {
			Serial.print("R: ");
			Serial.print(r);
			Serial.print("; ");
			Serial.println(state_right);
		}
		#endif
	}

	// read LEFT channel level
	if (level_left.available()) {
		float l = level_left.read();
		state_left = false;
		if (channel1) {
			if (channel1->UseAudio())
				state_left |= (l >= VOX_THRESHOLD);
			#ifdef ENABLE_SERIAL_KEYING
			if (channel1->UseDTR())
				state_left |= (Serial.dtr() ? true : false);
			if (channel1->UseRTS())
				state_left |= (Serial.rts() ? true : false);
			#endif

			#if defined (CLIP_LED_PIN) && (CLIP_LED_PIN >= 0)
			if (l >= channel1->ClipWarning()) {
				clip_found = true;
			}
			#endif

			// update the keying module
			channel1->SetState(state_left);
		}

		#ifdef VERBOSE_DEBUG
		if (l >= 0.01) {
			Serial.print("L: ");
			Serial.print(l);
			Serial.print("; ");
			Serial.println(state_left);
		}
		#endif
	}


	//
	//  Handle CLIP LED
	//
	#if defined (CLIP_LED_PIN) && (CLIP_LED_PIN >= 0)
	if (clip_found) {
		// light the CLIP LED
		digitalWrite(CLIP_LED_PIN, HIGH);
		lastClip = now;
	}

	if (Elapsed(now, lastClip) >= CLIP_DELAY) {
		// extinguish CLIP LED
		digitalWrite(CLIP_LED_PIN, LOW);
	}
	#endif
} // loop()

// EOF
